#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <dirent.h>
#include <unistd.h>

typedef struct {
	const char *name;
	int signal;
} SIG;

SIG signals[] = {
	{"HUP",  SIGHUP},
	{"INT",  SIGINT},
	{"QUIT", SIGQUIT},
	{"ILL",  SIGILL},
	{"TRAP", SIGTRAP},
	{"ABRT", SIGABRT},
	{"IOT",  SIGIOT},
	{"BUS",  SIGBUS},
	{"FPE",  SIGFPE},
	{"KILL", SIGKILL},
	{"USR1", SIGUSR1},
	{"SEGV", SIGSEGV},
	{"USR2", SIGUSR2},
	{"PIPE", SIGPIPE},
	{"ALRM", SIGALRM},
	{"TERM", SIGTERM},
	{"CHLD", SIGCHLD},
	{"STOP", SIGSTOP}
};


int parse_sig(char *arg) {
	int sig = atoi(arg);
	if (sig > 0 && sig <= NSIG)
		return sig;

	if (!strncasecmp(arg, "SIG", 3))
		arg += 3;

	for (size_t i = 0; i < sizeof(signals) / sizeof(signals[0]); i++)
		if (!strcasecmp(arg, signals[i].name))
			return signals[i].signal;

	return SIGTERM;
}

int my_kill(pid_t pid, int signal) {
	if (kill(pid, signal)) {
		fprintf(stderr, "kill: %d: %s\n", pid, strerror(errno));
		return 1;
	}

	return 0;
}

int main(int argc, char **argv) {
	unsigned int a_flag = 0;
	int signal = SIGTERM;

	int opt;
	while ((opt = getopt(argc, argv, "s:la")) != -1) {
		switch (opt) {
			case 'l':
				for (size_t i = 0; i < sizeof(signals) / sizeof(signals[0]); i++)
					printf("%zu) %s\n", i, signals[i].name);

				return 0;

			case 's':
				signal = parse_sig(optarg);
				break;

			case 'a':
				a_flag = 1;
				break;

			default:
				printf("kill [lsa] [pid1 pid2...]\n\t-s SIG\n\t-l List all signals\n\t-a Kill all process\n");
				return 0;
		}
	}

	argv += optind;
	argc -= optind;

	int ret = 0;

	if (a_flag) {
		DIR *dp = opendir("/proc");
		if (dp == NULL) {
			fprintf(stderr, "kill: /proc: %s\n", strerror(errno));
			return 1;
		}

		pid_t my_pid = getpid();

		struct dirent *ep;
		while ((ep = readdir(dp)) != NULL) {
			pid_t pid = atoi(ep->d_name);
			if (pid == 0 || pid == my_pid)
				continue;

			if (my_kill(pid, signal))
				ret = 1;
		}

		closedir(dp);
	}


	else {
		for (int i = 0; i < argc; i++) {
			pid_t pid = atoi(argv[i]);
			if (my_kill(pid, signal))
				ret = 1;
		}
	}

	return ret;
}
