#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <dirent.h>
#include <limits.h>
#include <errno.h>
#include <ctype.h>
#include <pwd.h>

int pscan(const char *pid) {
	char path[PATH_MAX + 1];

	/* Arguments */
	snprintf(path, sizeof(path), "/proc/%s/status", pid);
	int fd = open(path, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "ps: %s: %s\n", path, strerror(errno));
		return 1;
	}

	char status[2048];
	ssize_t n = read(fd, status, sizeof(status) - 1);
	status[n] = '\0';

	close(fd);

	char *prog = "none";
	char *state = "none";
	char *user = "none";

	/* Parsing */
	char *token = strtok(status, "\n");
	while (token) {
		char *val = strchr(token, ':');
		if (val == NULL) {
			fprintf(stderr, "ps: incorrect %s file\n", path);
			return 1;
		}

		*val = '\0';
		val++;

		/* Strip */
		while (isspace(*val))
			val++;

		/* Write values */
		if (!strncmp(token, "State", 5))
			state = val;

		else if (!strncmp(token, "Name", 4))
			prog = val;

		else if (!strncmp(token, "Uid", 3)) {
			struct passwd *pw = getpwuid(atoi(val));
			if (pw != NULL)
				user = pw->pw_name;
		}

		token = strtok(NULL, "\n");
	}

	printf("%s\t%s\t%s\t%s\n", pid, state, user, prog);
	return 0;
}

int main(int argc, char **argv) {

	int opt;
	while ((opt = getopt(argc, argv, "")) != -1) {
		puts("ps [PID]");
		return 0;
	}

	argv += optind;
	argc -= optind;

	printf("PID\tSTATE\tUSER\tCMD\n");

	int ret = 0;
	if (argc == 0) {
		DIR *dp = opendir("/proc");
		if (dp == NULL) {
			fprintf(stderr, "ps: /proc: %s\n", strerror(errno));
			return 1;
		}

		struct dirent *ep;
		while ((ep = readdir(dp)) != NULL)
			if (atoi(ep->d_name))
				if (pscan(ep->d_name))
					ret = 1;

		closedir(dp);
	}

	else {
		for (int i = 0; i < argc; i++)
			if (atoi(argv[i]))
				if (pscan(argv[i]))
					ret = 1;
	}

	return ret;
}
